unit uBuffer;

interface

uses
{$IF CompilerVersion <= 22}
  Forms, Classes, Windows, SysUtils, NetEncoding;
{$ELSE}
  Vcl.Forms, System.Classes, Winapi.Windows, System.SysUtils, System.NetEncoding;
{$ENDIF}

type
  TBuffer = class
  private
    function  GetDataLength: Integer; inline;
    procedure SetDataLength(Len: Integer); inline;
    function  GetItem(Index: Integer): Byte; inline;
    procedure SetItem(Index: Integer; Value: Byte); inline;
  public
    Data: TBytes;   //TBytes = array of Byte;
    procedure FromString(const Str: String); overload;  //ĬΪutf8
    procedure FromString(const Str: String; Encoding: TEncoding); overload;
    procedure FromHexString(const Str: String);
    procedure FromDelimitedHexString(HexStr: String; Prefix: String = '$'; Delimitor: String = ',');
    procedure FromBase64String(const Str: String);
    procedure FromBytes(const InBytes: TBytes; ByteLen: Integer = -1); overload;
    procedure FromBytes(const InBytes: array of Byte; ByteLen: Integer = -1); overload;
    procedure FromBytes(const Ints: array of UInt32; ByteLen: Integer); overload;
    procedure OffsetFromBytes(Offset: Integer; const Ints: array of UInt32; ByteLen: Integer);
    procedure FromStream(const Stream: TStream; ByteLen: Integer = -1);
    procedure FromFile(const FileName: String);
    function  ToString: String; reintroduce; overload;  //ĬΪutf8
    function  ToString(Encoding: TEncoding): String; reintroduce; overload;
    function  ToHexString: String;
    function  ToDelimitedHexString(Prefix: String = '$'; Delimitor: String = ', '): String;
    function  ToBase64String: String;
    procedure ToBytes(var OutBytes: TBytes; ByteLen: Integer = -1); overload;
    procedure ToBytes(var OutBytes: array of Byte; ByteLen: Integer = -1); overload;
    procedure ToBytes(var Ints: array of UInt32; ByteLen: Integer = 1); overload;
    procedure OffsetToBytes(Offset: Integer; var Ints: array of UInt32; ByteLen: Integer);
    procedure ToStream(const Stream: TStream);
    procedure ToFile(const FileName: String; Warning: Boolean = True);
    property  Length: Integer read GetDataLength write SetDataLength;
    property  Bytes[Index: Integer]: Byte read GetItem write SetItem; default;
  end;

resourcestring
  SInvalidBufSize = 'Invalid buffer size for ouput';

implementation

function TBuffer.GetDataLength: Integer;
begin
  Result := System.Length(Data);
end;

procedure TBuffer.SetDataLength(Len: Integer);
begin
  System.SetLength(Data, Len);
end;

function  TBuffer.GetItem(Index: Integer): Byte;
begin
  Result := Data[Index];
end;

procedure TBuffer.SetItem(Index: Integer; Value: Byte);
begin
  Data[Index] := Value;
end;

procedure TBuffer.FromString(const Str: String);
begin
  Data := TEncoding.UTF8.GetBytes(Str);
end;

procedure TBuffer.FromString(const Str: String; Encoding: TEncoding);
begin
  Data := Encoding.GetBytes(Str);
end;

procedure TBuffer.FromHexString(const Str: String);
var
  Len: Integer;
begin
  Len := System.Length(Str) div 2;
  SetLength(Data, Len);
  HexToBin(PChar(Str), @Data[0], Len)
end;

procedure TBuffer.FromDelimitedHexString(HexStr: String; Prefix: String; Delimitor: String);
var
  Len: Integer;
begin
  HexStr := HexStr.Replace(Prefix      , '');
  HexStr := HexStr.Replace(Delimitor , '');
  HexStr := HexStr.Replace(' '       , '');
  Len := System.Length(HexStr) div 2;
  SetLength(Data, Len);
  HexToBin(PChar(HexStr), @Data[0], Len)
end;

procedure TBuffer.FromBase64String(const Str: String);
var
  Base64Encoding: TBase64Encoding;
begin
//Base64Encoding := TBase64Encoding.Create; //з
  Base64Encoding := TBase64Encoding.Create(0); //з
  Data := Base64Encoding.DecodeStringToBytes(Str);
  Base64Encoding.Free;
end;

procedure TBuffer.FromBytes(const InBytes: array of Byte; ByteLen: Integer);
begin
  if (ByteLen = -1) then  ByteLen := System.Length(InBytes);
  SetLength(Data, ByteLen);
  Move(InBytes[0], Data[0], ByteLen);
end;

procedure TBuffer.FromBytes(const InBytes: TBytes; ByteLen: Integer);
begin
  if (ByteLen = -1) then  ByteLen := System.Length(InBytes);
  SetLength(Data, ByteLen);
  Move(InBytes[0], Data[0], ByteLen);
end;

procedure TBuffer.FromBytes(const Ints: array of UInt32; ByteLen: Integer);
begin
  SetLength(Data, ByteLen);
  Move(Ints[0], Data[0], ByteLen);
end;

procedure TBuffer.OffsetFromBytes(Offset: Integer; const Ints: array of UInt32; ByteLen: Integer);
begin
  Move(Ints[0], Data[Offset], ByteLen);
end;

procedure TBuffer.FromStream(const Stream: TStream; ByteLen: Integer);
begin
  if (ByteLen = -1) then ByteLen := Stream.Size;
  SetLength(Data, ByteLen);
  Stream.Read(Data, ByteLen);
end;

procedure TBuffer.FromFile(const FileName: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  SetLength(Data, Stream.Size);
  Stream.Read(Data, Stream.Size);
  Stream.Free;
end;

function TBuffer.ToString: String;
begin
  Result := TEncoding.UTF8.GetString(Data);
end;

function TBuffer.ToString(Encoding: TEncoding): String;
begin
  Result := Encoding.GetString(Data);
end;

function TBuffer.ToHexString: String;
var
  Len: Integer;
begin
  Len := System.Length(Data);
  SetLength(Result, 2*Len);
  BinToHex(@Data[0], PChar(Result), Len);
end;

function TBuffer.ToDelimitedHexString(Prefix: String; Delimitor: String): String;
var
  I, Len: Integer;
begin
  Result := '';
  Len := System.Length(Data);
  for I := 0 to Len-1 do
  begin
    Result := Result + Prefix + IntToHex(Data[I], 2);
    if I < Len-1 then
      Result := Result + Delimitor;
  end;
end;

function TBuffer.ToBase64String: String;
var
  Base64Encoding: TBase64Encoding;
begin
//Base64Encoding := TBase64Encoding.Create; //з
  Base64Encoding := TBase64Encoding.Create(0); //з
  Result := Base64Encoding.EncodeBytesToString(Data);
  Base64Encoding.Free;
end;

procedure TBuffer.ToBytes(var OutBytes: array of Byte; ByteLen: Integer);
begin
  if (ByteLen = -1) then ByteLen := System.Length(Data);
  if (ByteLen > System.Length(OutBytes)) then
    raise Exception.Create(SInvalidBufSize);
  Move(Data[0], OutBytes[0], ByteLen);
end;

procedure TBuffer.ToBytes(var OutBytes: TBytes; ByteLen: Integer);
begin
  if ByteLen = -1 then ByteLen := System.Length(Data);
  SetLength(OutBytes, ByteLen);
  Move(Data[0], OutBytes[0], ByteLen);
end;

procedure TBuffer.ToBytes(var Ints: array of UInt32; ByteLen: Integer);
begin
  if ByteLen = -1 then ByteLen := System.Length(Data);
  Move(Data[0], Ints[0], ByteLen);
end;

procedure TBuffer.OffsetToBytes(Offset: Integer; var Ints: array of UInt32; ByteLen: Integer);
begin
  Move(Data[Offset], Ints[0], ByteLen);
end;

procedure TBuffer.ToStream(const Stream: TStream);
begin
  Stream.Write(Data, System.Length(Data));
end;

procedure TBuffer.ToFile(const FileName: String; Warning: Boolean);
var
  Stream: TFileStream;
begin
  if Warning and FileExists(FileName) and
     (Application.MessageBox(PChar('File ' + FileName + ' Exists, Overwrite It?'),
                  'Warning: File Exists', MB_YESNO) = IDNO)  then Exit;

  Stream := TFileStream.Create(FileName, fmCreate);
  Stream.Write(Data, System.Length(Data));
  Stream.Free;
end;

end.
