unit uSM4;

interface

uses
{$IF CompilerVersion <= 22}
  Forms, Classes, Windows, SysUtils, NetEncoding,
{$ELSE}
  Vcl.Forms, System.Classes, Winapi.Windows, System.SysUtils, System.NetEncoding,
{$ENDIF}
  uConst, uBuffer;

type
  TCipherMode  = (cmECB, cmCBC, cmPCBC, cmCFB, cmOFB, cmCTR);
  TPaddingMode = (pmZERO, pmPKCS5, pmPKCS7, pmISO10126, pmANSIX923, pmOneAndZero);

  TSM4 = class(TObject)
  private type
    TWord = UInt32;
    TBlock    = array[0..4 -1] of TWord;
    TRoundKey = array[0..32-1] of TWord;
    TCryptType = (ctEncrypt, ctDecrypt);
    TProc = procedure(var A: TBlock) of object;
  private
    RK: TRoundKey;
    MK, DataBlock, IVBlock: TBlock;
    BlockSize, BlockLen, RKLen: Integer;
    function SBoxMap(A: TWord): TWord;
    function ROTL(A: TWord; N: Byte): TWord;
    procedure CryptBlock(const CryptType: TCryptType; var A: TBlock);
    procedure ExpandKey;
    procedure ReverseEndian(var A: TBlock);
    procedure AppendPadding;
    procedure RemovePadding;
    procedure CheckKeyBufferAndIVBuffer;
    procedure XorBlock(var A: TBlock; const B: TBlock); overload;
    procedure XorBlock(var A: TBlock; const B, C: TBlock); overload;
    procedure IncBlock(Var A: TBlock);
    procedure Crypt(CryptType: TCryptType);
  public
    CipherMode : TCipherMode;
    PaddingMode: TPaddingMode;
    SrcBuffer, KeyBuffer, IVBuffer, DestBuffer: TBuffer;
    procedure Encrypt;
    procedure Decrypt;
    constructor Create(aCipherMode: TCipherMode = cmECB; aPaddingMode: TPaddingMode = pmPKCS7);
    destructor Destroy; override;
  end;

implementation

constructor TSM4.Create(aCipherMode: TCipherMode; aPaddingMode: TPaddingMode);
begin
  inherited Create;
  CipherMode  := aCipherMode;
  PaddingMode := aPaddingMode;
  SrcBuffer  := TBuffer.Create;
  KeyBuffer  := TBuffer.Create;
  IVBuffer   := TBuffer.Create;
  DestBuffer := TBuffer.Create;
  BlockSize := SizeOf(TBlock);  //16
  BlockLen  := SizeOf(TBlock) div Sizeof(TWord);  //4
  RKLen := SizeOf(TRoundKey) div Sizeof(TWord); //32
end;

destructor TSM4.Destroy;
begin
  SrcBuffer.Free;
  KeyBuffer.Free;
  IVBuffer.Free;
  DestBuffer.Free;
  inherited;
end;

procedure TSM4.Encrypt;
begin
  AppendPadding;
  Crypt(ctEncrypt);
end;

procedure TSM4.Decrypt;
begin
  Crypt(ctDecrypt);
  RemovePadding;
end;

procedure TSM4.Crypt(CryptType: TCryptType);
var
  InBlock: TBlock;
  DataLen, OffSet: Integer;
begin
  CheckKeyBufferAndIVBuffer;
  DataLen := SrcBuffer.Length;
  DestBuffer.Length := DataLen;
  KeyBuffer.ToBytes(MK, BlockSize);
  if CipherMode <> cmECB then
  begin
    IVBuffer.ToBytes(IVBlock, BlockSize);
  end;
  ExpandKey;
  OffSet := 0;
  while OffSet < DataLen do
  begin
    SrcBuffer.OffsetToBytes(OffSet, DataBlock, BlockSize);
    case CipherMode of
      cmECB: //Electronic Codebook (ECB)
        begin
          CryptBlock(CryptType, DataBlock);
        end;
      cmCBC: //Cipher Block Chaining (CBC)
        begin
          if CryptType = ctEncrypt then
          begin
            XorBlock(DataBlock, IVBlock);
            CryptBlock(CryptType, DataBlock);
            IVBlock := DataBlock;
          end
          else
          begin
            InBlock := DataBlock;
            CryptBlock(CryptType, DataBlock);
            XorBlock(DataBlock, IVBlock);
            IVBlock := InBlock;
          end;
        end;
      cmPCBC: //Propagating Cipher Block Chaining (PCBC)
        begin
          InBlock := DataBlock;
          if CryptType = ctEncrypt then
          begin
            XorBlock(DataBlock, IVBlock);
            CryptBlock(CryptType, DataBlock);
          end
          else
          begin
            CryptBlock(CryptType, DataBlock);
            XorBlock(DataBlock, IVBlock);
          end;
          XorBlock(IVBlock, DataBlock, InBlock);
        end;
      cmCFB: //Cipher Feedback (CFB)
        begin
          InBlock := DataBlock;
          CryptBlock(ctEncrypt, IVBlock); //ǼܻǽܣﶼEncryptIVв
          XorBlock(DataBlock, IVBlock);
          if CryptType = ctEncrypt then
            IVBlock := DataBlock
          else
            IVBlock  := InBlock;
        end;
      cmOFB: //Output Feedback (OFB)
        begin
          CryptBlock(ctEncrypt, IVBlock); //ǼܻǽܣﶼEncryptIVв
          XorBlock(DataBlock, IVBlock);
        end;
      cmCTR: //Counter (CTR)
        begin
          InBlock := IVBlock;
          CryptBlock(ctEncrypt, InBlock); //ǼܻǽܣﶼEncryptIVв
          XorBlock(DataBlock, InBlock);
          IncBlock(IVBlock);
        end;
    end;
    DestBuffer.OffsetFromBytes(OffSet, DataBlock, BlockSize);
    Inc(OffSet, BlockSize);
  end;
end;

procedure TSM4.AppendPadding; //
var
  I, M, N, Len, NewLen: Integer;
begin
  Len := Length(SrcBuffer.Data);
  M := (Len mod BlockSize);
  if  (M = 0) and (PaddingMode = pmZero) then
    N := 0
  else
    N := BlockSize - M;
  NewLen := Len + N;
  SetLength(SrcBuffer.Data, NewLen);

  case PaddingMode of
    pmPKCS5, pmPKCS7:
    begin
      for I := Len to NewLen-1 do SrcBuffer.Data[I] := N;
    end;
    pmANSIX923:
    begin
      for I := Len to NewLen-1 do
      begin
        if I < NewLen-1 then
          SrcBuffer.Data[I] := 0
        else
          SrcBuffer.Data[I] := N;
      end;
    end;
    pmISO10126:
    begin
      Randomize;
      for I := Len to NewLen-1 do
      begin
        if I < NewLen-1 then
          SrcBuffer.Data[I] := Random(255)
        else
          SrcBuffer.Data[I] := N;
      end;
    end;
    pmZERO:
    begin
      for I := Len to NewLen-1 do SrcBuffer.Data[I] := 0;
    end;
    pmOneAndZero:
    begin
      for I := Len to NewLen-1 do
      begin
        if I = Len then
          SrcBuffer.Data[I] := $80
        else
          SrcBuffer.Data[I] := 0;
      end;
    end;
  end;
end;

procedure TSM4.RemovePadding; //Ƴ
var
  I, M, Len: Integer;
begin
  Len := Length(DestBuffer.Data);
  case PaddingMode of
    pmPKCS5, pmPKCS7, pmANSIX923, pmISO10126:
    begin
      M := DestBuffer.Data[Len-1];
      SetLength(DestBuffer.Data,  Len-M);
    end;
    pmZERO:
    begin
      for I := Len-1 downto 0 do
      begin
        if DestBuffer.Data[I] = 0 then
          Dec(Len)
        else
          Break;
      end;
      SetLength(DestBuffer.Data, Len);
    end;
    pmOneAndZero:
    begin
      for I := Len-1 downto 0 do
      begin
        if DestBuffer.Data[I] <> $80 then
          Dec(Len)
        else
          Break;
      end;
      SetLength(DestBuffer.Data, Len-1);
    end;
  end;
end;

procedure TSM4.CheckKeyBufferAndIVBuffer; //KeyIV16Bytes0
var
  I, Len: Integer;
begin
  Len := Length(KeyBuffer.Data);
  SetLength(KeyBuffer.Data, BlockSize);
  for I := Len to BlockSize - 1 do
  begin
    KeyBuffer.Data[I] := 0;
  end;

  if CipherMode <> cmECB then //ECBģʽҪʼʸIV
  begin
    Len := Length(IVBuffer.Data);
    SetLength(IVBuffer.Data, BlockSize);
    for I := Len to BlockSize - 1 do
    begin
      IVBuffer.Data[I] := 0;
    end;
  end;
end;

procedure TSM4.ExpandKey;
  function Transform(A: TWord): TWord; //ϳû
  begin
    A := SBoxMap(A); //Ա任
    Result := A xor ROTL(A, 13) xor ROTL(A, 23); //Ա任
  end;
var
  K: TBlock;
  I: Integer;
  M0, M1, M2, M3: TWord;
begin
  ReverseEndian(MK); //32λСת

  for I := 0 to BlockLen-1 do
  begin
    K[I] := MK[I] xor FK[I];
  end;

  for I := 0 to RKLen-1 do //32ε
  begin
    M0 := (I  ) mod 4;
    M1 := (I+1) mod 4;
    M2 := (I+2) mod 4;
    M3 := (I+3) mod 4;
    K[M0] := K[M0] xor Transform(K[M1] xor K[M2] xor K[M3] xor CK[I]); //
    RK[I] := K[M0];
  end;
end;

procedure TSM4.CryptBlock(const CryptType: TCryptType; var A: TBlock);
  function TransForm(A: TWord): TWord; //ϳû
  begin
    A := SBoxMap(A); //Ա任
    Result := A xor ROTL(A, 2) xor ROTL(A, 10) xor ROTL(A, 18) xor ROTL(A, 24); //Ա任
  end;
var
  T: TBlock;
  I,  R: Integer;
  M0, M1, M2, M3: TWord;
begin
  ReverseEndian(A); //32λС(Endian)ת

  for I := 0 to RKLen-1 do  //32ε
  begin
    if CryptType = ctEncrypt then
      R := I
    else
      R := 31 - I;  //ʱʹ÷RK

    M0 := (I  ) mod 4;
    M1 := (I+1) mod 4;
    M2 := (I+2) mod 4;
    M3 := (I+3) mod 4;

    A[M0] := A[M0] xor TransForm(A[M1] xor A[M2] xor A[M3] xor RK[R]); //
  end;

  ReverseEndian(A); //32λС(Endian)ת

  T := A;
  for I := 0 to 3 do //Blockڲ432λλ任
  begin
    A[I] := T[3 - I];
  end;
end;

function TSM4.SBoxMap(A: TWord): TWord; //SзԱ任(Byte to ByteĹ̶ӳ}
var
  I: Integer;
  B: array[0..3] of Byte absolute A;
  R: array[0..3] of Byte absolute Result;
begin
  for I := 0 to 3 do
  begin
    R[I] := SBox[B[I]];
  end;
end;

procedure TSM4.ReverseEndian(var A: TBlock); //32λС(Endian)任
var
  I: Integer;
begin
  for I := 0 to BlockLen-1 do
  begin
    A[I] := ((A[I] and $FF000000) shr 24) or
            ((A[I] and $00FF0000) shr 8 ) or
            ((A[I] and $0000FF00) shl 8 ) or
            ((A[I] and $000000FF) shl 24);
  end;
end;

function TSM4.ROTL(A: TWord; N: Byte): TWord; //32λѭNλ
begin
  Result := (A shl N) or (A shr (32-N));
end;

procedure TSM4.XorBlock(var A: TBlock; const B: TBlock); //Block
var
  I: Integer;
begin
  for I := 0 to BlockLen-1 do
  begin
    A[I] := A[I] xor B[I];
  end;
end;

procedure TSM4.XorBlock(var A: TBlock; const B, C: TBlock); //Block
var
  I: Integer;
begin
  for I := 0 to BlockLen-1 do
  begin
    A[I] := B[I] xor C[I];
  end;
end;

procedure TSM4.IncBlock(var A: TBlock); //Blockֵ1
var
  I: Integer;
begin
  for I := 0 to BlockLen-1 do
  begin
    if I > 0 then
    begin
      A[I-1] := 0;
    end;
    if (I = BlockLen-1) or (A[I] < $FFFFFFFF) then
    begin
      Inc(A[0]);
      Exit;
    end;
  end;
end;

end.
